/*! ********************************************
 * SMFSequencer.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.smf
{
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;

public final class SMFSequencer implements IGrayBoxModule
{
// EventBuffer -----
	private var _eventBuffer     :Vector.<MidiEvent> = null;
	private var _eventBufferSize :int = 256;
	private var _usedEvents      :int = 0;
	public function clearMIDIEventBuffer():void
	{
		for (var i:int = 0; i < _eventBufferSize; i++)_eventBuffer[i].clear();
		_usedEvents = 0;
	}
	public function getEventBufferSize():int { return _eventBufferSize; }
	public function setEventBufferSize(size:int):int
	{
		if (size == _eventBufferSize) return _eventBufferSize;
		_eventBufferSize = size;
		_eventBuffer.fixed = false;
		_eventBuffer.length = size;
		_eventBuffer.fixed = true;
		for (var i:int = 0; i < size; i++)
		{
			if (_eventBuffer[i] == null) _eventBuffer[i] = new MidiEvent();
			_eventBuffer[i].clear();
		}
		_usedEvents = 0;
		return _eventBufferSize;
	}
// Sequence -----
	private var _sequence  :SMFSequence = null;
	private var _track     :SMFTrackEvent = null; // Event List - "MTrk" Events
	private var _division  :int = 480;  // "MThd" divison
	private var _loopable  :Boolean = false;
	private var _loopStart :SMFTrackEvent = null;
	private var _loopEnd   :SMFTrackEvent = null;
	public function getSequence():SMFSequence{ return _sequence; }
	public function setSequence(seq:SMFSequence):Boolean
	{
		initialize();
		if (seq.track == null) return false;
		// ---
		_sequence = seq;
		_track = seq.track;
		_division = seq.division;
		_loopable = seq.loopable;
		_loopStart = seq.loopStart;
		_loopEnd = seq.loopEnd;
		// ---
		_refTime = _division * 1000000.0 / _sampleRate;
		_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
		gotoHead();
		return true;
	}
	public function loopable():Boolean { return _loopable; }
// Sequencer -----
	//public var midiInPorts  :int = 0xFFFF;
	public var midiOutPorts :int = 0xFFFF;
	/*
	   if (event.type <= _eventFilter) --> process
	   0xE0 = ChannelEvent Only
	   0xFF = include SysEx
	   0xFFFF = include SysEx and MetaEvent
	*/
	private var _eventFilter :int = 0xFF;
	private var _sampleRate :Number = 44100.0;
	private var _tempo      :Number = 500000.0;  // SetTempo (MetaEvent FF51)
	private var _isPlay     :Boolean = false;
	private var _loopPlay   :Boolean = false;
	private var _trackPos   :SMFTrackEvent = null;
	// event delay timer
	private const _EV_TIMER_FPBITS   :int = 24;
	private const _EV_TIMER_FPVALUE  :int = 1 << _EV_TIMER_FPBITS;
	private const _EV_TIMER_FPMASK   :int = _EV_TIMER_FPVALUE - 1;
	private const _EV_TIMER_FPVALUE_N:Number = Number(_EV_TIMER_FPVALUE);
	private var _refTime     :Number = 0.0;  // val = _timeDivision(Note Ticks) * (1000000uSec / SampleRate)
	private var _evTimer     :int = 0;
	private var _evTimerAFP  :int = 0;
	private var _evTimerStep :int = 0;
	
// Sequencer status ---
	public function getSampleRate():Number { return _sampleRate; }
	public function setSampleRate(val:Number):void
	{ 
		_sampleRate = val;
		_refTime = _division * 1000000.0 / _sampleRate;
		_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	}
	public function getOutputMetaMessage():Boolean { return _eventFilter == 0xFFFF; }
	public function setOutputMetaMessage(flag:Boolean):void { _eventFilter = (flag)? 0xFFFF : 0xFF; }
	public function playable():Boolean { return _track != null; }
	public function isPlay():Boolean { return _isPlay; }
	public function isLoopPlay():Boolean { return _loopPlay; }
	public function endOfTrack():Boolean { return _trackPos == null; }
// Sequencer Contorols ---
	public function reset():void
	{
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = null;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
		gotoHead();
	}
	public function initialize():void
	{
		// data ---
		_sequence = null;
		_track = null;
		_division = 480;
		_loopable = false;
		_loopStart = null;
		_loopEnd = null;
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = null;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
	}
	public function getBPM():Number { return 60000000.0 / _tempo; }
	public function setBPM(val:Number):void
	{ 
		_tempo = 60000000.0 / val;
		_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	}
	public function play(resume:Boolean = false):Boolean
	{
		if (_track == null) return false;
		if (!resume || _trackPos == null)gotoHead();
		return _isPlay = true;
	}
	public function stop(pause:Boolean = false):void
	{
		_isPlay = false;
		if (_track == null) return;
		if (!pause)gotoHead();
	}
	public function loopPlay(flag:Boolean):Boolean
	{
		if (_loopable) return _loopPlay = flag;
		else return false;
	}
	public function gotoHead():void
	{
		if (_track == null) return;
		_trackPos = _track;
		// ---
		_tempo = 500000.0;
		_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
		// ---
		_evTimer = _trackPos.deltaTime;
		_evTimerAFP = 0;
	}
// Interface - IfGrayBoxModule ---
	public function updateSampleRate(sampleRate:Number):void{ setSampleRate(sampleRate); }
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	
	public function resetModule():void{ clearMIDIEventBuffer(); reset(); }
	public function initiModule():void{ clearMIDIEventBuffer(); initialize(); }
	public function connectModule(host:GrayBox):void
	{ 
		clearMIDIEventBuffer();
		setSampleRate(host.getSampleRate());
	}
	public function disconnectModule():void{ clearMIDIEventBuffer(); }
	public function updateBypassStatus(bypass:Boolean):void{ clearMIDIEventBuffer(); }
	
	public function startProcess():void{ clearMIDIEventBuffer(); }
	public function stopProcess():void{ clearMIDIEventBuffer(); }
	public function suspendProcess():void{ clearMIDIEventBuffer(); }
	public function resumeProcess():void{ clearMIDIEventBuffer(); }
	public function process(procData:ProcessData):void
	{
		if (!_isPlay) return;
		_usedEvents = 0;
		var i:int, _ev:MidiEvent,
			_list:MidiEvent = procData.midiEventList,
			_sampleFrames:int = procData.sampleFrames;
		for (i = 0; i < _sampleFrames; i++)
		{
			while (_evTimer <= 0)
			{
			// Arbitrary Loop -->
				if (_trackPos == _loopEnd)
				{
					if (_loopPlay)
					{
						_trackPos = _loopStart;
						//if (callBackLoop != null) callBackLoop();
						//continue;
					}
				}
			// Process Event -->
				if (_trackPos.type <= _eventFilter)
				{
					if (_usedEvents < _eventBufferSize)
					{
						_ev = _eventBuffer[_usedEvents++];
						_ev.next = null;  // clear
						_ev.midiPorts = midiOutPorts;
						_ev.offsetFrames = i;
						_ev.type = _trackPos.type;
						_ev.channel = _trackPos.channel;
						_ev.value0 = _trackPos.value0;
						_ev.value1 = _trackPos.value1;
						_ev.data = _trackPos.data;
					}
				}
			// Tempo -->
				if (_trackPos.type == 0xFF51)
				{
					_tempo = _trackPos.value0;
					_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
				}
			// NextEvent, EOT -->
				_trackPos = _trackPos.next;
				if (_trackPos != null)_evTimer += _trackPos.deltaTime;
				else 
				{
					if (_loopPlay) { gotoHead(); continue; }
					else { _isPlay = false; _evTimer = 1; i = _sampleFrames; }
				}
			}
			// Counter -----
			_evTimerAFP += _evTimerStep;
			_evTimer -= _evTimerAFP >> _EV_TIMER_FPBITS;
			_evTimerAFP &= _EV_TIMER_FPMASK;
		}
		// addtoList -->
		for (i = 0; i < _usedEvents; i++)
		{
			_ev = _eventBuffer[i];
			while (_list.next != null)
			{
				if (_ev.offsetFrames >= _list.next.offsetFrames) _list = _list.next;
				else break;
			}
			_ev.next = _list.next;
			_list.next = _ev;
		}
	}
// Constructor ---
	public function SMFSequencer(eventBufferSize:int = 512):void
	{
		_eventBufferSize = eventBufferSize;
		_eventBuffer = new Vector.<MidiEvent>(_eventBufferSize, true);
		for (var i:int = 0; i < _eventBufferSize; i++) _eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
	}
}
}